import uuid
import logging
from datetime import datetime
from typing import List, Union, Tuple

from sirius_sdk.agent.pairwise import Pairwise
from sirius_sdk.agent.wallet.abstract.crypto import AbstractCrypto
from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.agent.microledgers import Transaction, Microledger, MicroledgerList
from sirius_sdk.agent.sm import AbstractStateMachine, StateMachineTerminatedWithError
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack, Status
from sirius_sdk.agent.consensus.simple.messages import *

# Problem codes
REQUEST_NOT_ACCEPTED = "request_not_accepted"
REQUEST_PROCESSING_ERROR = 'request_processing_error'
RESPONSE_NOT_ACCEPTED = "response_not_accepted"
RESPONSE_PROCESSING_ERROR = 'response_processing_error'


class MicroLedgerSimpleConsensus(AbstractStateMachine):

    def __init__(
            self, crypto: AbstractCrypto, me: Pairwise.Me,
            pairwise_list: AbstractPairwiseList, microledgers: MicroledgerList, *args, **kwargs
    ):
        self.__me = me
        self.__problem_report = None
        self.__microledgers = microledgers
        self.__crypto = crypto
        self.__pairwise_list = pairwise_list
        self.__transport = None
        self.__thread_id = None
        self.__cached_p2p = {}
        super().__init__(*args, **kwargs)

    @property
    def me(self) -> Pairwise.Me:
        return self.__me

    @property
    def crypto(self) -> AbstractCrypto:
        return self.__crypto

    @property
    def microledgers(self) -> MicroledgerList:
        return self.__microledgers

    @property
    def pairwise_list(self) -> AbstractPairwiseList:
        return self.__pairwise_list

    @property
    def protocols(self) -> List[str]:
        return [SimpleConsensusMessage.PROTOCOL, Ack.PROTOCOL]

    @property
    def problem_report(self) -> SimpleConsensusProblemReport:
        return self.__problem_report

    async def init_microledger(
            self, ledger_name: str, participants: List[str], genesis: List[Transaction]
    ) -> (bool, Microledger):
        participants = list(set(participants + [self.me.did]))
        await self._setup('simple-consensus-' + uuid.uuid4().hex, self.time_to_live)
        try:
            await self.log(progress=0, message=f'Create ledger [{ledger_name}]')
            ledger, genesis = await self.microledgers.create(ledger_name, genesis)
            await self.log(message=f'Ledger creation terminated successfully')
            try:
                await self._init_microledger_internal(ledger, participants, genesis)
                await self.log(progress=100, message='All participants accepted ledger creation')
            except Exception as e:
                await self.microledgers.reset(ledger_name)
                await self.log(message=f'Reset ledger')
                if isinstance(e, StateMachineTerminatedWithError):
                    await self.log(
                        progress=100, message=f'Terminated with error',
                        problem_code=e.problem_code, explain=e.explain
                    )
                    return False, None
                else:
                    await self.log(
                        progress=100, message=f'Terminated with exception',
                        exception=str(e)
                    )
                    raise
            else:
                return True, ledger
        finally:
            await self._clean()

    async def accept_microledger(self, actor: Pairwise, propose: InitRequestLedgerMessage) -> (bool, Microledger):
        if self.me.did not in propose.participants:
            raise SiriusContextError('Invalid state machine initialization')
        time_to_live = propose.timeout_sec or self.time_to_live
        await self._setup(propose.thread_id, time_to_live)
        try:
            ledger_name = propose.ledger.get('name', None)
            try:
                if not ledger_name:
                    await self._terminate_with_problem_report(
                        problem_code=REQUEST_PROCESSING_ERROR,
                        explain='Ledger name is Empty!',
                        their_did=actor.their.did,
                    )
                for their_did in propose.participants:
                    if their_did != self.me.did:
                        pw = await self.get_p2p(their_did, raise_exception=False)
                        if pw is None:
                            await self._terminate_with_problem_report(
                                problem_code=REQUEST_PROCESSING_ERROR,
                                explain=f'Pairwise for DID: "their_did" does not exists!' % their_did,
                                their_did=actor.their.did
                            )
                await self.log(progress=0, message=f'Start ledger [{ledger_name}] creation process')
                ledger = await self._accept_microledger_internal(actor, propose, time_to_live)
                await self.log(progress=100, message='Ledger creation terminated successfully')
            except Exception as e:
                await self.microledgers.reset(ledger_name)
                await self.log(message=f'Reset ledger')
                if isinstance(e, StateMachineTerminatedWithError):
                    await self.log(
                        progress=100, message=f'Terminated with error',
                        problem_code=e.problem_code, explain=e.explain
                    )
                    return False, None
                else:
                    await self.log(
                        progress=100, message=f'Terminated with exception',
                        exception=str(e)
                    )
                    raise
            else:
                return True, ledger
        finally:
            await self._clean()

    async def commit(
            self, ledger: Microledger, participants: List[str], transactions: List[Transaction]
    ) -> (bool, Optional[List[Transaction]]):
        await self._setup('simple-consensus-txns-' + uuid.uuid4().hex, self.time_to_live)
        try:
            try:
                await self.log(progress=0, message=f'Start committing {len(transactions)} transactions')
                txns = await self._commit_internal(ledger, transactions, participants)
                await self.log(progress=100, message='Commit operation was accepted by all participants')
                return True, txns
            except Exception as e:
                await ledger.reset_uncommitted()
                await self.log(message='Reset uncommitted')
                if isinstance(e, StateMachineTerminatedWithError):
                    await self.log(
                        progress=100, message=f'Terminated with error',
                        problem_code=e.problem_code, explain=e.explain
                    )
                    return False, None
                else:
                    await self.log(
                        progress=100, message=f'Terminated with exception',
                        exception=str(e)
                    )
                    raise
        finally:
            await self._clean()

    async def accept_commit(self, actor: Pairwise, propose: ProposeTransactionsMessage) -> bool:
        time_to_live = propose.timeout_sec or self.time_to_live
        await self._setup(propose.thread_id, time_to_live)
        try:
            ledger = None
            try:
                await self.log(progress=0, message=f'Start acception {len(propose.transactions)} transactions')
                ledger = await self._load_ledger(actor, propose)
                await self._accept_commit_internal(ledger, actor, propose)
                await self.log(progress=100, message='Acception terminated successfully')
                return True
            except Exception as e:
                if ledger:
                    await ledger.reset_uncommitted()
                    await self.log(message='Reset uncommitted')
                if isinstance(e, StateMachineTerminatedWithError):
                    await self.log(
                        progress=100, message=f'Terminated with error',
                        problem_code=e.problem_code, explain=e.explain
                    )
                    await self.log(
                        progress=100, message=f'Terminated with exception',
                        exception=str(e)
                    )
                    return False
                else:
                    raise
        finally:
            await self._clean()

    async def _switch(
            self, their_did: str, req: Union[SimpleConsensusMessage, SimpleConsensusProblemReport, Ack]
    ) -> (bool, Union[SimpleConsensusMessage, SimpleConsensusProblemReport]):
        if isinstance(req, SimpleConsensusMessage) or isinstance(req, SimpleConsensusProblemReport) or isinstance(req, Ack):
            pw = await self.get_p2p(their_did)
            self.__transport.pairwise = pw
            ok, resp = await self.__transport.switch(req)
            if isinstance(resp, SimpleConsensusMessage) or isinstance(resp, SimpleConsensusProblemReport) or isinstance(resp, Ack):
                return ok, resp
            else:
                await self._terminate_with_problem_report(
                    problem_code=RESPONSE_NOT_ACCEPTED,
                    explain=f'Unexpected message with @type = {resp.type}',
                    their_did=their_did
                )
        else:
            raise SiriusContextError('Unexpected req type')

    async def _send(self, their_did: Union[str, List[str]], message: AriesProtocolMessage):
        if isinstance(their_did, str):
            pw = await self.get_p2p(their_did)
            self.__transport.pairwise = pw
            await self.__transport.send(message)
        elif isinstance(their_did, list):
            tos = [await self.get_p2p(did) for did in their_did]
            await self.__transport.send_many(message, tos)
        else:
            raise SiriusContextError('Unexpected their_did type')

    async def _switch_multiple(
            self, their_did: List[str], message: AriesProtocolMessage
    ) -> List[Tuple[bool, AriesProtocolMessage]]:
        tos = [await self.get_p2p(did) for did in their_did]
        vk2did = {p.their.verkey: p.their.did for p in tos}
        responses = await self.__transport.send_many(message, tos)
        results = []
        await_list = {}
        for did, response in zip(their_did, responses):
            success, body = response
            item = (success, None)
            results.append(item)
            if success:
                await_list[did] = None
        if await_list:
            for n in range(len(await_list)):
                message, sender_verkey, recipient_verkey = await self.__transport.get_one()
                did = vk2did[sender_verkey]
                await_list[did] = message
            for i, did in enumerate(their_did):
                success, body = results[i]
                if success is True:
                    msg = await_list.get(did, None)
                    results[i] = (success, msg)
            return results
        else:
            return []

    async def _setup(self, thread_id: str, time_to_live: int):
        self.__thread_id = thread_id
        self.__transport = await self.transports.spawn(thread_id)
        await self.__transport.start(time_to_live=time_to_live)

    async def _clean(self):
        if self.__transport:
            await self.__transport.stop()
        self.__transport = None

    async def get_p2p(self, their_did: str, raise_exception: bool = True) -> Pairwise:
        if their_did not in self.__cached_p2p.keys():
            pw = await self.pairwise_list.load_for_did(their_did)
            if pw is None and raise_exception:
                raise SiriusContextError('Pairwise for "%s" does not exists!' % their_did)
            self.__cached_p2p[their_did] = pw
        return self.__cached_p2p[their_did]

    async def _terminate_with_problem_report(
            self, problem_code: str, explain, their_did: Union[str, List[str]], raise_exception: bool = True
    ):
        self.__problem_report = SimpleConsensusProblemReport(
            problem_code=problem_code, explain=explain, thread_id=self.__thread_id
        )
        if isinstance(their_did, str):
            await self._send(their_did, self.__problem_report)
        elif isinstance(their_did, list):
            await self._send(their_did, self.__problem_report)
        else:
            raise SiriusContextError('Unexpected their_did type')
        if raise_exception:
            raise StateMachineTerminatedWithError(problem_code, explain)

    async def _init_microledger_internal(
            self, ledger: Microledger, participants: List[str], genesis: List[Transaction]
    ):
        # ============= STAGE 1: PROPOSE =================
        propose = InitRequestLedgerMessage(
            timeout_sec=self.time_to_live,
            ledger_name=ledger.name,
            genesis=genesis,
            root_hash=ledger.root_hash,
            participants=[did for did in participants]
        )
        await propose.add_signature(self.crypto, self.me)
        request_commit = InitResponseLedgerMessage()
        request_commit.assign_from(propose)
        neighbours = [did for did in participants if did != self.me.did]

        await self.log(progress=20, message='Send propose', payload=dict(propose))
        results = await self._switch_multiple(neighbours, propose)
        await self.log(progress=30, message='Received responses from all participants')
        neighbours_responses = {did: msg for (ok, msg), did in zip(results, neighbours) if ok is True}
        if len(neighbours) != len(neighbours):
            error_neighbours = [did for (ok, _), did in zip(results, neighbours) if not ok]
            await self._terminate_with_problem_report(
                problem_code=REQUEST_PROCESSING_ERROR,
                explain='Stage-1: Participants [%s] unreachable' % ','.join(error_neighbours),
                their_did=neighbours
            )

        await self.log(progress=40, message='Validate responses')
        for their_did, response in neighbours_responses.items():
            if isinstance(response, InitResponseLedgerMessage):
                response.validate()
                await response.check_signatures(self.crypto, their_did)
                signature = response.signature(their_did)
                request_commit.signatures.append(signature)
            elif isinstance(response, SimpleConsensusProblemReport):
                self.__problem_report = response
                logging.error('Code: %s; Explain: %s' % (response.problem_code, response.explain))
                raise StateMachineTerminatedWithError(response.problem_code, response.explain)

        # ============= STAGE 2: COMMIT ============
        await self.log(progress=60, message='Send commit request', payload=dict(request_commit))
        results = await self._switch_multiple(neighbours, request_commit)
        await self.log(progress=70, message='Received commit responses')
        neighbours_responses = {did: msg for (ok, msg), did in zip(results, neighbours) if ok is True}

        await self.log(progress=80, message='Validate commit responses from neighbours')
        acks = []
        for their_did, response in neighbours_responses.items():
            if isinstance(response, SimpleConsensusProblemReport):
                neighbours = [did for did in neighbours if did != their_did]
                await self._terminate_with_problem_report(response.problem_code, response.explain, neighbours)
            else:
                acks.append(their_did)
        # ============== STAGE 3: POST-COMMIT ============
        if set(acks) == set(neighbours):
            ack = Ack(thread_id=self.__thread_id, status=Status.OK)
            await self.log(progress=90, message='All checks OK. Send Ack to neighbors')
            await self._send(neighbours, ack)
        else:
            acks_str = ','.join(acks)
            neighbours_str = ','.join(neighbours)
            await self._terminate_with_problem_report(
                REQUEST_PROCESSING_ERROR,
                f'Stage-3: Actual list of acceptors: [{acks_str}]  Expected: [{neighbours_str}]',
                neighbours
            )

    async def _accept_microledger_internal(
            self, actor: Pairwise, propose: InitRequestLedgerMessage, timeout: int
    ) -> Microledger:
        # =============== STAGE 1: PROPOSE ===============
        try:
            propose.validate()
            await propose.check_signatures(self.crypto, actor.their.did)
            if len(propose.participants) < 2:
                raise SiriusValidationError('Stage-1: participants less than 2')
        except SiriusValidationError as e:
            await self._terminate_with_problem_report(REQUEST_NOT_ACCEPTED, e.message, actor.their.did)
        genesis = [Transaction(txn) for txn in propose.ledger['genesis']]
        await self.log(progress=10, message='Initialize ledger')
        ledger, txns = await self.microledgers.create(propose.ledger['name'], genesis)
        await self.log(progress=20, message='Ledger initialized successfully')
        if propose.ledger['root_hash'] != ledger.root_hash:
            await self.microledgers.reset(ledger.name)
            await self._terminate_with_problem_report(
                REQUEST_PROCESSING_ERROR, 'Stage-1: Non-consistent Root Hash', actor.their.did
            )
        response = InitResponseLedgerMessage(timeout_sec=timeout)
        response.assign_from(propose)
        commit_ledger_hash = response.ledger_hash
        await response.add_signature(self.crypto, self.me)
        # =============== STAGE 2: COMMIT ===============
        await self.log(progress=30, message='Send propose response', payload=dict(response))
        ok, request_commit = await self._switch(actor.their.did, response)
        if ok:
            await self.log(progress=50, message='Validate request commit')
            if isinstance(request_commit, InitResponseLedgerMessage):
                try:
                    request_commit.validate()
                    hashes = await request_commit.check_signatures(self.crypto, participant='ALL')
                    for their_did, decoded in hashes.items():
                        if decoded != commit_ledger_hash:
                            raise SiriusValidationError(f'Stage-2: NonEqual Ledger hash with participant "{their_did}"')
                except SiriusValidationError as e:
                    await self._terminate_with_problem_report(REQUEST_NOT_ACCEPTED, e.message, actor.their.did)
                commit_participants_set = set(request_commit.participants)
                propose_participants_set = set(propose.participants)
                signers_set = set([s['participant'] for s in request_commit.signatures])
                if propose_participants_set != signers_set:
                    error_explain = 'Stage-2: Set of signers differs from proposed participants set'
                elif commit_participants_set != signers_set:
                    error_explain = 'Stage-2: Set of signers differs from commit participants set'
                else:
                    error_explain = None
                if error_explain:
                    await self._terminate_with_problem_report(
                        problem_code=REQUEST_NOT_ACCEPTED,
                        explain=error_explain,
                        their_did=actor.their.did
                    )
                else:
                    # Accept commit
                    await self.log(progress=70, message='Send Ack')
                    ack = Ack(thread_id=self.__thread_id, status=Status.OK)
                    ok, resp = await self._switch(actor.their.did, ack)
                    # =========== STAGE-3: POST-COMMIT ===============
                    if ok:
                        await self.log(progress=90, message='Response to Ack received')
                        if isinstance(resp, Ack):
                            return ledger
                        elif isinstance(resp, SimpleConsensusProblemReport):
                            self.__problem_report = resp
                            logging.error(
                                'Code: %s; Explain: %s' % (resp.problem_code, resp.explain))
                            raise StateMachineTerminatedWithError(
                                self.__problem_report.problem_code, self.__problem_report.explain
                            )
                    else:
                        await self._terminate_with_problem_report(
                            problem_code=REQUEST_PROCESSING_ERROR,
                            explain='Stage-3: Commit accepting was terminated by timeout for actor: %s' % actor.their.did,
                            their_did=actor.their.did
                        )
            elif isinstance(request_commit, SimpleConsensusProblemReport):
                self.__problem_report = request_commit
                raise StateMachineTerminatedWithError(
                    self.__problem_report.problem_code, self.__problem_report.explain
                )
        else:
            await self._terminate_with_problem_report(
                problem_code=REQUEST_PROCESSING_ERROR,
                explain='Stage-2: Commit response awaiting was terminated by timeout for actor: %s' % actor.their.did,
                their_did=actor.their.did
            )

    async def _commit_internal(
            self, ledger: Microledger, transactions: List[Transaction], participants: List[str]
    ) -> List[Transaction]:
        participants = list(set(participants + [self.me.did]))
        neighbours = [did for did in participants if did != self.me.did]
        for their_did in neighbours:
            await self.get_p2p(their_did, raise_exception=True)
        verkeys = {}
        for their_did in neighbours:
            pw = await self.get_p2p(their_did)
            verkeys[their_did] = pw.their.verkey
        txn_time = str(datetime.utcnow())
        start, end, txns = await ledger.append(transactions, txn_time)
        propose = ProposeTransactionsMessage(
            transactions=txns,
            state=MicroLedgerState.from_ledger(ledger),
            participants=participants,
            timeout_sec=self.time_to_live
        )
        # ==== STAGE-1 Propose transactions to participants ====
        commit = CommitTransactionsMessage(participants=participants)
        self_pre_commit = PreCommitTransactionsMessage(state=propose.state)
        await self_pre_commit.sign_state(self.crypto, self.me)
        commit.add_pre_commit(
            participant=self.me.did,
            pre_commit=self_pre_commit
        )
        awaited_list = []
        await self.log(progress=20, message='Send Propose to participants', payload=dict(propose))
        results = await self._switch_multiple(neighbours, propose)
        await self.log(progress=30, message='Received Propose from participants')
        neighbours_responses = {did: msg for (ok, msg), did in zip(results, neighbours) if ok is True}
        if len(neighbours) != len(neighbours):
            error_neighbours = [did for (ok, _), did in zip(results, neighbours) if not ok]
            await self._terminate_with_problem_report(
                problem_code=REQUEST_PROCESSING_ERROR,
                explain='Stage-1: Participants [%s] unreachable' % ','.join(error_neighbours),
                their_did=neighbours
            )

        await self.log(progress=50, message='Validate responses')
        for their_did, pre_commit in neighbours_responses.items():
            if isinstance(pre_commit, PreCommitTransactionsMessage):
                try:
                    pre_commit.validate()
                    success, state = await pre_commit.verify_state(self.crypto, verkeys[their_did])
                    if not success:
                        raise SiriusValidationError(
                            f'Stage-1: Error verifying signed ledger state for participant {their_did}'
                        )
                    if pre_commit.hash != propose.state.hash:
                        raise SiriusValidationError(
                            f'Stage-1: Non-consistent ledger state for participant {their_did}'
                        )
                except SiriusValidationError as e:
                    await self._terminate_with_problem_report(
                        RESPONSE_NOT_ACCEPTED,
                        f'Stage-1: Error for participant {their_did}: "{e.message}"',
                        neighbours
                    )
                else:
                    commit.add_pre_commit(their_did, pre_commit)
                    awaited_list.append(their_did)
            elif isinstance(pre_commit, SimpleConsensusProblemReport):
                explain = f'Stage-1: Problem report from participant {their_did} "{pre_commit.explain}"'
                self.__problem_report = SimpleConsensusProblemReport(pre_commit.problem_code, explain)
                await self._send([did for did in neighbours if did != their_did], self.__problem_report)
                raise StateMachineTerminatedWithError(
                    self.__problem_report.problem_code, self.__problem_report.explain
                )

        # ===== STAGE-2: Accumulate pre-commits and send commit propose to all participants
        post_commit_all = PostCommitTransactionsMessage()
        await post_commit_all.add_commit_sign(self.crypto, commit, self.me)
        await self.log(progress=60, message='Send Commit to participants', payload=dict(commit))
        results = await self._switch_multiple(neighbours, commit)
        await self.log(progress=70, message='Received Commit response from participants')
        neighbours_responses = {did: msg for (ok, msg), did in zip(results, neighbours) if ok is True}
        if len(neighbours) != len(neighbours):
            error_neighbours = [did for (ok, _), did in zip(results, neighbours) if not ok]
            await self._terminate_with_problem_report(
                problem_code=REQUEST_PROCESSING_ERROR,
                explain='Stage-1: Participants [%s] unreachable' % ','.join(error_neighbours),
                their_did=neighbours
            )
        await self.log(progress=80, message='Validate responses')
        for their_did, post_commit in neighbours_responses.items():
            if isinstance(post_commit, PostCommitTransactionsMessage):
                try:
                    post_commit.validate()
                except SiriusValidationError as e:
                    await self._terminate_with_problem_report(
                        RESPONSE_NOT_ACCEPTED,
                        f'Stage-2: Error for participant {their_did}: "{e.message}"',
                        neighbours
                    )
                else:
                    post_commit_all['commits'].extend(post_commit.commits)
            elif isinstance(post_commit, SimpleConsensusProblemReport):
                explain = f'Stage-2: Problem report from participant {their_did} "{post_commit.explain}"'
                self.__problem_report = SimpleConsensusProblemReport(post_commit.problem_code, explain)
                await self._send([did for did in neighbours if did != their_did], self.__problem_report)
                raise StateMachineTerminatedWithError(
                    self.__problem_report.problem_code, self.__problem_report.explain
                )

        # ===== STAGE-3: Notify all participants with post-commits and finalize process
        await self.log(progress=90, message='Send Post-Commit', payload=dict(post_commit_all))
        await self._send(neighbours, post_commit_all)
        uncommitted_size = ledger.uncommitted_size - ledger.size
        await ledger.commit(uncommitted_size)
        return txns

    async def _load_ledger(self, actor: Pairwise, propose: ProposeTransactionsMessage) -> Microledger:
        neighbours = [did for did in propose.participants if did != self.me.did]
        try:
            propose.validate()
            if len(propose.participants) < 2:
                raise SiriusValidationError(f'Stage-1: participant count less than 2')
            if self.me.did not in propose.participants:
                raise SiriusValidationError(f'Stage-1: {self.me.did} is not participant')
            for their_did in neighbours:
                pw = await self.get_p2p(their_did)
                if pw is None:
                    raise SiriusValidationError(f'Stage-1: Pairwise for did: {their_did} does not exists')
            is_ledger_exists = await self.microledgers.is_exists(propose.state.name)
            if not is_ledger_exists:
                raise SiriusValidationError(f'Stage-1: Ledger with name {propose.state.name} does not exists')
        except SiriusValidationError as e:
            await self._terminate_with_problem_report(
                problem_code=RESPONSE_NOT_ACCEPTED,
                explain=e.message,
                their_did=actor.their.did
            )
        ledger = await self.microledgers.ledger(propose.state.name)
        return ledger

    async def _accept_commit_internal(self, ledger: Microledger, actor: Pairwise, propose: ProposeTransactionsMessage):
        neighbours = [did for did in propose.participants if did != self.me.did]
        # ===== STAGE-1: Process Propose, apply transactions and response ledger state on self-side
        await ledger.append(propose.transactions)
        ledger_state = MicroLedgerState.from_ledger(ledger)
        pre_commit = PreCommitTransactionsMessage(state=MicroLedgerState.from_ledger(ledger))
        await pre_commit.sign_state(self.crypto, self.me)
        await self.log(progress=10, message='Send Pre-Commit', payload=dict(pre_commit))
        ok, commit = await self._switch(actor.their.did, pre_commit)
        if ok:
            await self.log(progress=20, message='Received Pre-Commit response', payload=dict(commit))
            if isinstance(commit, CommitTransactionsMessage):
                # ===== STAGE-2: Process Commit request, check neighbours signatures
                try:
                    await self.log(progress=30, message='Validate Commit')
                    if set(commit.participants) != set(propose.participants):
                        raise SiriusValidationError('Non-consistent participants')
                    commit.validate()
                    await commit.verify_pre_commits(self.crypto, ledger_state)
                except SiriusValidationError as e:
                    await self._terminate_with_problem_report(
                        problem_code=REQUEST_NOT_ACCEPTED,
                        explain=f'Stage-2: error for actor {actor.their.did}: "{e.message}"',
                        their_did=actor.their.did
                    )
                else:
                    # ===== STAGE-3: Process post-commit, verify participants operations
                    post_commit = PostCommitTransactionsMessage()
                    await post_commit.add_commit_sign(self.crypto, commit, self.me)

                    await self.log(progress=50, message='Send Post-Commit', payload=dict(post_commit))
                    ok, post_commit_all = await self._switch(actor.their.did, post_commit)
                    if ok:
                        await self.log(progress=60, message='Received Post-Commit response', payload=dict(post_commit_all))
                        if isinstance(post_commit_all, PostCommitTransactionsMessage):
                            try:
                                await self.log(progress=80, message='Validate response')
                                post_commit_all.validate()
                                verkeys = [(await self.get_p2p(did)).their.verkey for did in neighbours]
                                await post_commit_all.verify_commits(self.crypto, commit, verkeys)
                            except SiriusValidationError as e:
                                await self._terminate_with_problem_report(
                                    problem_code=REQUEST_NOT_ACCEPTED,
                                    explain=f'Stage-3: error for actor {actor.their.did}: "{e.message}"',
                                    their_did=actor.their.did
                                )
                            else:
                                uncommitted_size = ledger_state.uncommitted_size - ledger_state.size
                                await self.log(progress=90, message='Flush transactions to Ledger storage')
                                await ledger.commit(uncommitted_size)
                        elif isinstance(post_commit_all, SimpleConsensusProblemReport):
                            explain = f'Stage-3: Problem report from actor {actor.their.did}: "{post_commit_all.explain}"'
                            self.__problem_report = SimpleConsensusProblemReport(post_commit_all.problem_code, explain)
                            raise StateMachineTerminatedWithError(
                                self.__problem_report.problem_code, self.__problem_report.explain
                            )
                    else:
                        await self._terminate_with_problem_report(
                            problem_code=REQUEST_PROCESSING_ERROR,
                            explain=f'Stage-3: Post-Commit awaiting terminated by timeout for actor: {actor.their.did}',
                            their_did=actor.their.did
                        )
            elif isinstance(commit, SimpleConsensusProblemReport):
                explain = f'Stage-1: Problem report from actor {actor.their.did}: "{commit.explain}"'
                self.__problem_report = SimpleConsensusProblemReport(commit.problem_code, explain)
                raise StateMachineTerminatedWithError(
                    self.__problem_report.problem_code, self.__problem_report.explain
                )
            else:
                await self._terminate_with_problem_report(
                    REQUEST_NOT_ACCEPTED, 'Unexpected message @type: %s' % str(commit.type), actor.their.did
                )
        else:
            await self._terminate_with_problem_report(
                problem_code=REQUEST_PROCESSING_ERROR,
                explain=f'Stage-1: Commit awaiting terminated by timeout for actor: {actor.their.did}',
                their_did=actor.their.did
            )



import uuid
from datetime import datetime

import pytest

from sirius_sdk import Agent
from sirius_sdk.agent.microledgers import Transaction, LedgerMeta


@pytest.mark.asyncio
async def test_init_ledger(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"}
        ]
        ledger, txns = await agent4.microledgers.create(ledger_name, genesis_txns)
        assert ledger.root_hash == '3u8ZCezSXJq72H5CdEryyTuwAKzeZnCZyfftJVFr7y8U'
        assert all([isinstance(txn, Transaction) for txn in txns])
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_merkle_info(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"},
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op5"},
        ]
        ledger, txns = await agent4.microledgers.create(ledger_name, genesis_txns)
        merkle_info = await ledger.merkle_info(4)
        assert merkle_info.root_hash == 'CwX1TRYKpejHmdnx3gMgHtSioDzhDGTASAD145kjyyRh'
        assert merkle_info.audit_path == ['46kxvYf7RjRERXdS56vUpFCzm2A3qRYSLaRr6tVT6tSd', '3sgNJmsXpmin7P5C6jpHiqYfeWwej5L6uYdYoXTMc1XQ']
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_append_operations(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"}
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)
        txns = [
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"},
        ]
        txn_time = str(datetime.now())
        start, end, appended_transactions = await ledger.append(txns, txn_time)
        assert end == 3
        assert start == 2
        assert all([isinstance(txn, Transaction) for txn in appended_transactions])
        assert txn_time in str(appended_transactions)
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_commit_discard(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"}
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)
        txns = [
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"},
        ]
        txn_time = str(datetime.now())

        assert ledger.uncommitted_root_hash == ledger.root_hash
        await ledger.append(txns, txn_time)
        assert ledger.uncommitted_root_hash != ledger.root_hash
        assert ledger.size == 1
        assert ledger.uncommitted_size == 3

        # COMMIT
        await ledger.commit(1)
        assert ledger.size == 2
        assert ledger.uncommitted_size == 3
        assert ledger.uncommitted_root_hash != ledger.root_hash

        # DISCARD
        await ledger.discard(1)
        assert ledger.size == 2
        assert ledger.uncommitted_size == 2
        assert ledger.uncommitted_root_hash == ledger.root_hash

    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_reset_uncommitted(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"}
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)
        txns = [
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"},
        ]
        await ledger.append(txns)
        uncommitted_size_before = ledger.uncommitted_size
        await ledger.reset_uncommitted()
        uncommitted_size_after = ledger.uncommitted_size

        assert uncommitted_size_after != uncommitted_size_before
        assert uncommitted_size_after == 1
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_get_operations(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"}
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)
        txns = [
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op5"},
        ]
        await ledger.append(txns)

        # 1 get_last_committed_txn
        txn = await ledger.get_last_committed_transaction()
        assert isinstance(txn, Transaction)
        assert 'op3' in str(txn)

        # 2 get_last_txn
        txn = await ledger.get_last_transaction()
        assert isinstance(txn, Transaction)
        assert 'op5' in str(txn)

        # 3 get_uncommitted_txns
        txns = await ledger.get_uncommitted_transactions()
        assert all([isinstance(txn, Transaction) for txn in txns])
        assert all(op in str(txns) for op in ['op4', 'op5']) is True
        assert any(op in str(txns) for op in ['op1', 'op2', 'op3']) is False

        # 4 get_by_seq_no
        txn = await ledger.get_transaction(seq_no=1)
        assert isinstance(txn, Transaction)
        assert 'op1' in str(txn)

        # 5 get_by_seq_no_uncommitted
        txn = await ledger.get_uncommitted_transaction(seq_no=4)
        assert isinstance(txn, Transaction)
        assert 'op4' in str(txn)
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_reset(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"},
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op5"},
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)
        assert ledger.size == 5

        is_exists = await agent4.microledgers.is_exists(ledger_name)
        assert is_exists is True

        await agent4.microledgers.reset(ledger_name)
        is_exists = await agent4.microledgers.is_exists(ledger_name)
        assert is_exists is False
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_list(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"}
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)
        # Get List
        collection = await agent4.microledgers.list()
        assert ledger_name in str(collection)
        assert all([isinstance(meta, LedgerMeta) for meta in collection])

        # Is exists
        ok = await agent4.microledgers.is_exists(ledger_name)
        assert ok is True

        # Reset Calling
        await agent4.microledgers.reset(ledger_name)

        # Get List
        collection = await agent4.microledgers.list()
        assert ledger_name not in str(collection)
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_get_all_txns(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"}
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)
        txns = [
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op5"},
        ]
        await ledger.append(txns)

        txns = await ledger.get_all_transactions()
        assert all([isinstance(txn, Transaction) for txn in txns])
        assert all(op in str(txns) for op in ['op4', 'op5']) is False
        assert any(op in str(txns) for op in ['op1', 'op2', 'op3']) is True
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_audit_proof(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"},
            {"reqId": 3, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op3"},
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op5"},
            {"reqId": 6, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op6"},
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)
        txns = [
            {"reqId": 7, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op7"},
            {"reqId": 8, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op8"},
            {"reqId": 9, "identifier": "CECeGXDi6EHuhpwz19uyjjEnsRGNXodFYqCRgdLmLRkt", "op": "op9"},
        ]
        await ledger.append(txns)

        audit_paths = []
        for seq_no in [1, 2, 3, 4, 5, 6]:
            audit_proof = await ledger.audit_proof(seq_no)
            assert audit_proof.root_hash == '3eDS4j8HgpAyRnuvfFG624KKvQBuNXKBenhqHmHtUgeq'
            assert audit_proof.ledger_size == 6
            assert audit_proof.audit_path not in audit_paths
            audit_paths.append(audit_proof.audit_path)

        for seq_no in [7, 8, 9]:
            audit_proof = await ledger.audit_proof(seq_no)
            assert audit_proof.root_hash == '3eDS4j8HgpAyRnuvfFG624KKvQBuNXKBenhqHmHtUgeq'
            assert audit_proof.ledger_size == 6
            audit_paths.append(audit_proof.audit_path)
            assert ledger.uncommitted_root_hash == 'Dkoca8Af15uMLBHAqbddwqmpiqsaDEtKDoFVfNRXt44g'
        print('@')
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_leaf_hash(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"}
        ]
        ledger, txns = await agent4.microledgers.create(ledger_name, genesis_txns)
        txn = txns[0]
        leaf_hash = await agent4.microledgers.leaf_hash(txn)
        assert isinstance(leaf_hash, bytes)
        expected = b'y\xd9\x92\x9f\xd1\xe7\xf1o\t\x9c&\xb6\xf4HP\xda\x04J\xd0\xfeQ\xe9.X-\x9c\xa3r\xf2\xb8\xb90'
        assert expected == leaf_hash
    finally:
        await agent4.close()


@pytest.mark.asyncio
async def test_rename(agent4: Agent, ledger_name: str):
    await agent4.open()
    try:
        genesis_txns = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"}
        ]
        ledger, _ = await agent4.microledgers.create(ledger_name, genesis_txns)

        new_name = 'new_name_' + uuid.uuid4().hex
        await ledger.rename(new_name)
        assert ledger.name == new_name

        is_exists1 = await agent4.microledgers.is_exists(ledger_name)
        is_exists2 = await agent4.microledgers.is_exists(new_name)
        assert is_exists1 is False
        assert is_exists2 is True
    finally:
        await agent4.close()



from sirius_sdk.agent.agent import Agent
from sirius_sdk.encryption import P2PConnection
from sirius_sdk.agent.pairwise import Pairwise, TheirEndpoint
from sirius_sdk.agent.connections import Endpoint


__all__ = ["Agent", "P2PConnection", "Pairwise", "TheirEndpoint", "Endpoint"]



from abc import ABC, abstractmethod
from typing import List, Optional
from urllib.parse import urlparse, urlunparse

from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.agent.wallet.abstract.pairwise import AbstractPairwise


class TheirEndpoint:

    def __init__(self, endpoint: str, verkey: str, routing_keys: List[str]=None):
        self.endpoint = endpoint
        self.verkey = verkey
        self.routing_keys = routing_keys or []

    @property
    def netloc(self) -> Optional[str]:
        if self.endpoint:
            return urlparse(self.endpoint).netloc
        else:
            return None

    @netloc.setter
    def netloc(self, value: str):
        if self.endpoint:
            components = list(urlparse(self.endpoint))
            components[1] = value
            self.endpoint = urlunparse(components)


class Pairwise:

    class Their(TheirEndpoint):

        def __init__(self, did: str, label: str, endpoint: str, verkey: str, routing_keys: List[str]=None):
            self.did = did
            self.label = label
            super().__init__(endpoint, verkey, routing_keys)

    class Me:

        def __init__(self, did, verkey):
            self.did = did
            self.verkey = verkey

        def __eq__(self, other):
            if isinstance(other, Pairwise.Me):
                return self.did == other.did and self.verkey == other.verkey

    def __init__(self, me: Me, their: Their, metadata: dict=None):
        self.__me = me
        self.__their = their
        self.__metadata = metadata

    @property
    def their(self) -> Their:
        return self.__their

    @property
    def me(self) -> Me:
        return self.__me

    @property
    def metadata(self) -> dict:
        return self.__metadata


class AbstractPairwiseList(ABC):

    @abstractmethod
    async def create(self, pairwise: Pairwise):
        raise NotImplemented

    @abstractmethod
    async def update(self, pairwise: Pairwise):
        raise NotImplemented

    @abstractmethod
    async def is_exists(self, their_did: str) -> bool:
        raise NotImplemented

    @abstractmethod
    async def ensure_exists(self, pairwise: Pairwise):
        raise NotImplemented

    @abstractmethod
    async def load_for_did(self, their_did: str) -> Optional[Pairwise]:
        raise NotImplemented

    @abstractmethod
    async def load_for_verkey(self, their_verkey: str) -> Optional[Pairwise]:
        raise NotImplemented


class WalletPairwiseList(AbstractPairwiseList):

    def __init__(self, api: (AbstractPairwise, AbstractDID)):
        self._api_pairwise = api[0]
        self._api_did = api[1]

    async def create(self, pairwise: Pairwise):
        await self._api_did.store_their_did(did=pairwise.their.did, verkey=pairwise.their.verkey)
        await self._api_pairwise.create_pairwise(
            their_did=pairwise.their.did,
            my_did=pairwise.me.did,
            metadata=pairwise.metadata,
            tags=self._build_tags(pairwise)
        )

    async def update(self, pairwise: Pairwise):
        await self._api_pairwise.set_pairwise_metadata(
            their_did=pairwise.their.did,
            metadata=pairwise.metadata,
            tags=self._build_tags(pairwise)
        )

    async def is_exists(self, their_did: str) -> bool:
        return await self._api_pairwise.is_pairwise_exists(their_did=their_did)

    async def ensure_exists(self, pairwise: Pairwise):
        if await self.is_exists(their_did=pairwise.their.did):
            await self.update(pairwise)
        else:
            await self.create(pairwise)

    async def load_for_did(self, their_did: str) -> Optional[Pairwise]:
        if await self.is_exists(their_did):
            raw = await self._api_pairwise.get_pairwise(their_did)
            metadata = raw['metadata']
            pairwise = self._restore_pairwise(metadata)
            return pairwise
        else:
            return None

    async def load_for_verkey(self, their_verkey: str) -> Optional[Pairwise]:
        collection, count = await self._api_pairwise.search(tags={'their_verkey': their_verkey}, limit=1)
        if collection:
            metadata = collection[0]['metadata']
            pairwise = self._restore_pairwise(metadata)
            return pairwise
        else:
            return None

    @staticmethod
    def _build_tags(p: Pairwise):
        return {
            'my_did': p.me.did,
            'my_verkey': p.me.verkey,
            'their_verkey': p.their.verkey
        }

    @staticmethod
    def _restore_pairwise(metadata: dict):
        pairwise = Pairwise(
            me=Pairwise.Me(
                did=metadata.get('me', {}).get('did', None),
                verkey=metadata.get('me', {}).get('verkey', None)
            ),
            their=Pairwise.Their(
                did=metadata.get('their', {}).get('did', None),
                verkey=metadata.get('their', {}).get('verkey', None),
                label=metadata.get('their', {}).get('label', None),
                endpoint=metadata.get('their', {}).get('endpoint', {}).get('address', None),
                routing_keys=metadata.get('their', {}).get('endpoint', {}).get('routing_keys', None),
            ),
            metadata=metadata
        )
        return pairwise



from typing import List, Union
from datetime import datetime, timedelta

from sirius_sdk.agent.pairwise import Pairwise
from sirius_sdk.agent.aries_rfc.utils import utc_to_str
from sirius_sdk.agent.wallet.abstract.anoncreds import AbstractAnonCreds
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache, CacheOptions
from sirius_sdk.agent.sm import AbstractStateMachine, StateMachineTerminatedWithError
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack, Status
from sirius_sdk.agent.aries_rfc.feature_0037_present_proof.messages import *


PROPOSE_NOT_ACCEPTED = "propose_not_accepted"
RESPONSE_NOT_ACCEPTED = "response_not_accepted"
RESPONSE_PROCESSING_ERROR = "response_processing_error"
REQUEST_NOT_ACCEPTED = "request_not_accepted"
RESPONSE_FOR_UNKNOWN_REQUEST = "response_for_unknown_request"
REQUEST_PROCESSING_ERROR = 'request_processing_error'
VERIFY_ERROR = 'verify_error'


class Verifier(AbstractStateMachine):

    def __init__(
            self, api: AbstractAnonCreds, cache: AbstractCache, prover: Pairwise, pool_name: str, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.__api = api
        self.__prover = prover
        self.__transport = None
        self.__cache = cache
        self.__problem_report = None
        self.__pool_name = pool_name

    async def verify(
            self, proof_request: dict, translation: List[AttribTranslation] = None,
            comment: str = None, locale: str = BasePresentProofMessage.DEF_LOCALE
    ):
        await self.__start()
        try:
            try:
                # Step-1: Send proof request
                expires_time = datetime.utcnow() + timedelta(seconds=self.time_to_live)
                request_msg = RequestPresentationMessage(
                    proof_request=proof_request,
                    translation=translation,
                    comment=comment,
                    locale=locale,
                    expires_time=utc_to_str(expires_time)
                )
                request_msg.please_ack = True
                presentation = await self.__switch(request_msg)
                if not isinstance(presentation, PresentationMessage):
                    raise StateMachineTerminatedWithError(
                        RESPONSE_NOT_ACCEPTED, 'Unexpected @type: %s' % str(presentation.type)
                    )
                # Step-2 Verify
                identifiers = presentation.proof.get('identifiers', [])
                schemas = {}
                credential_defs = {}
                rev_reg_defs = {}
                rev_regs = {}
                opts = CacheOptions()
                for identifier in identifiers:
                    schema_id = identifier['schema_id']
                    cred_def_id = identifier['cred_def_id']
                    rev_reg_id = identifier['rev_reg_id']
                    if schema_id and schema_id not in schemas:
                        schemas[schema_id] = await self.__cache.get_schema(
                            self.__pool_name, self.__prover.me.did, schema_id, opts
                        )
                    if cred_def_id and cred_def_id not in credential_defs:
                        credential_defs[cred_def_id] = await self.__cache.get_cred_def(
                            self.__pool_name, self.__prover.me.did, cred_def_id, opts
                        )
                success = await self.__api.verifier_verify_proof(
                    proof_request=proof_request,
                    proof=presentation.proof,
                    schemas=schemas,
                    credential_defs=credential_defs,
                    rev_reg_defs=rev_reg_defs,
                    rev_regs=rev_regs
                )
                if success:
                    ack = Ack(thread_id=presentation.id, status=Status.OK)
                    await self.__send(ack)
                    return True
                else:
                    raise StateMachineTerminatedWithError(VERIFY_ERROR, 'Verifying return false')
            except StateMachineTerminatedWithError as e:
                self.__problem_report = PresentProofProblemReport(e.problem_code, e.explain)
                if e.notify:
                    await self.__send(self.__problem_report)
                return False
        finally:
            await self.__stop()

    @property
    def problem_report(self) -> PresentProofProblemReport:
        return self.__problem_report

    @property
    def protocols(self) -> List[str]:
        return [BasePresentProofMessage.PROTOCOL, Ack.PROTOCOL]

    async def __start(self):
        self.__transport = await self.transports.spawn(self.__prover)
        await self.__transport.start(self.protocols, self.time_to_live)

    async def __stop(self):
        if self.__transport:
            await self.__transport.stop()
            self.__transport = None

    async def __switch(self, request: BasePresentProofMessage) -> Union[BasePresentProofMessage, Ack]:
        ok, resp = await self.__transport.switch(request)
        if ok:
            if isinstance(resp, BasePresentProofMessage) or isinstance(resp, Ack):
                try:
                    resp.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(RESPONSE_PROCESSING_ERROR, e.message)
                return resp
            elif isinstance(resp, PresentProofProblemReport):
                raise StateMachineTerminatedWithError(resp.problem_code, resp.explain, notify=False)
            else:
                raise StateMachineTerminatedWithError(RESPONSE_PROCESSING_ERROR, 'Unexpected response @type: %s' % str(resp.type))
        else:
            raise StateMachineTerminatedWithError(RESPONSE_PROCESSING_ERROR, 'Response awaiting terminated by timeout')

    async def __send(self, msg: Union[BasePresentProofMessage, Ack, PresentProofProblemReport]):
        await self.__transport.send(msg)


class Prover(AbstractStateMachine):

    def __init__(
            self, api: AbstractAnonCreds, cache: AbstractCache, verifier: Pairwise, pool_name: str, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.__api = api
        self.__verifier = verifier
        self.__transport = None
        self.__cache = cache
        self.__problem_report = None
        self.__pool_name = pool_name

    async def prove(self, request: RequestPresentationMessage, master_secret_id: str) -> bool:
        await self.__start()
        try:
            try:
                # Step-1: Process proof-request
                try:
                    request.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, e.message)
                cred_infos, schemas, credential_defs, rev_states = await self._extract_credentials_info(
                    request.proof_request, self.__pool_name
                )
                if cred_infos.get('requested_attributes', None) or cred_infos.get('requested_predicates', None):
                    # Step-2: Build proof
                    proof = await self.__api.prover_create_proof(
                        proof_req=request.proof_request,
                        requested_credentials=cred_infos,
                        master_secret_name=master_secret_id,
                        schemas=schemas,
                        credential_defs=credential_defs,
                        rev_states=rev_states
                    )
                    # Step-3: Send proof and wait Ack to check success from Verifier side
                    presentation_msg = PresentationMessage(proof)
                    presentation_msg.please_ack = True
                    # Step-3: Walt ACK
                    ack = await self.__switch(presentation_msg)
                    if isinstance(ack, Ack):
                        return True
                    else:
                        raise StateMachineTerminatedWithError(
                            RESPONSE_FOR_UNKNOWN_REQUEST, 'Unexpected response @type: %s' % str(ack.type)
                        )
                else:
                    raise StateMachineTerminatedWithError(
                        REQUEST_PROCESSING_ERROR, 'No proof correspondent to proof-request'
                    )
            except StateMachineTerminatedWithError as e:
                self.__problem_report = PresentProofProblemReport(e.problem_code, e.explain)
                if e.notify:
                    await self.__send(self.__problem_report)
                return False
        finally:
            await self.__stop()

    @property
    def problem_report(self) -> PresentProofProblemReport:
        return self.__problem_report

    @property
    def protocols(self) -> List[str]:
        return [BasePresentProofMessage.PROTOCOL, Ack.PROTOCOL]

    async def _extract_credentials_info(self, proof_request, pool_name: str) -> (dict, dict, dict, dict):
        # Extract credentials from wallet that satisfy to request
        proof_response = await self.__api.prover_search_credentials_for_proof_req(proof_request, limit_referents=1)
        schemas = {}
        credential_defs = {}
        rev_states = {}
        opts = CacheOptions()
        requested_credentials = {
            'self_attested_attributes': {},
            'requested_attributes': {},
            'requested_predicates': {}
        }
        all_infos = []
        for referent_id, cred_infos in proof_response['requested_attributes'].items():
            cred_info = cred_infos[0]['cred_info']  # Get first
            info = {
                'cred_id': cred_info['referent'],
                'revealed': True
            }
            requested_credentials['requested_attributes'][referent_id] = info
            all_infos.append(cred_info)
        for referent_id, predicates in proof_response['requested_predicates'].items():
            pred_info = predicates[0]['cred_info']  # Get first
            info = {
                'cred_id': pred_info['referent']
            }
            requested_credentials['requested_predicates'][referent_id] = info
            all_infos.append(pred_info)
        for cred_info in all_infos:
            schema_id = cred_info['schema_id']
            cred_def_id = cred_info['cred_def_id']
            schema = await self.__cache.get_schema(
                pool_name=pool_name, submitter_did=self.__verifier.me.did,
                id_=schema_id, options=opts
            )
            cred_def = await self.__cache.get_cred_def(
                pool_name=pool_name, submitter_did=self.__verifier.me.did,
                id_=cred_def_id, options=opts
            )
            schemas[schema_id] = schema
            credential_defs[cred_def_id] = cred_def
        return requested_credentials, schemas, credential_defs, rev_states

    async def __start(self):
        self.__transport = await self.transports.spawn(self.__verifier)
        await self.__transport.start(self.protocols, self.time_to_live)

    async def __stop(self):
        if self.__transport:
            await self.__transport.stop()
            self.__transport = None

    async def __switch(self, request: BasePresentProofMessage) -> Union[BasePresentProofMessage, Ack]:
        ok, resp = await self.__transport.switch(request)
        if ok:
            if isinstance(resp, BasePresentProofMessage) or isinstance(resp, Ack):
                try:
                    resp.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(REQUEST_PROCESSING_ERROR, e.message)
                return resp
            elif isinstance(resp, PresentProofProblemReport):
                raise StateMachineTerminatedWithError(resp.problem_code, resp.explain, notify=False)
            else:
                raise StateMachineTerminatedWithError(
                    REQUEST_PROCESSING_ERROR, 'Unexpected response @type: %s' % str(resp.type)
                )
        else:
            raise StateMachineTerminatedWithError(REQUEST_PROCESSING_ERROR, 'Response awaiting terminated by timeout')

    async def __send(self, msg: Union[BasePresentProofMessage, Ack, PresentProofProblemReport]):
        await self.__transport.send(msg)



import json
from typing import List, Union, Dict

from sirius_sdk.errors.exceptions import *
from sirius_sdk.agent.connections import AgentRPC


METADATA_ATTR = 'txnMetadata'


def serialize_ordering(value: dict) -> bytes:
    data = json.dumps(value, sort_keys=True, ensure_ascii=False, separators=(',', ':')).encode()
    return data


class Transaction(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if METADATA_ATTR not in self:
            self[METADATA_ATTR] = {}

    def has_metadata(self) -> bool:
        if METADATA_ATTR in self.keys():
            meta = self[METADATA_ATTR]
            return len(meta.keys()) > 0
        else:
            return False

    @staticmethod
    def create(*args, **kwargs):
        inst = Transaction(*args, **kwargs)
        if inst[METADATA_ATTR] != {}:
            raise SiriusContextError('"%s" attribute must be empty for new transaction' % METADATA_ATTR)
        return inst

    @staticmethod
    def from_value(from_: Union[List[dict], dict]) -> Union[List[Dict], Dict]:
        if isinstance(from_, list):
            return [Transaction(txn) for txn in from_]
        elif isinstance(from_, dict):
            return Transaction(from_)
        else:
            raise SiriusContextError('Unexpected input value')


class MerkleInfo:

    def __init__(self, root_hash: str, audit_path: List[str]):
        self.__root_hash = root_hash
        self.__audit_path = audit_path

    @property
    def root_hash(self) -> str:
        return self.__root_hash

    @property
    def audit_path(self) -> List[str]:
        return self.__audit_path


class LedgerMeta(dict):

    def __init__(self, name: str, uid: str, created: str):
        super().__init__()
        self['name'] = name
        self['uid'] = uid
        self['created'] = created

    @property
    def name(self) -> str:
        return self['name']

    @property
    def uid(self) -> str:
        return self['uid']

    @property
    def created(self) -> str:
        return self['created']


class AuditProof(MerkleInfo):

    def __init__(self, root_hash: str, audit_path: List[str], ledger_size: int):
        super().__init__(root_hash, audit_path)
        self.__ledger_size = ledger_size

    @property
    def ledger_size(self) -> int:
        return self.__ledger_size


class Microledger:

    def __init__(self, name: str, api: AgentRPC):
        self.__name = name
        self.__api = api
        self.__state = None

    @property
    def name(self) -> str:
        return self.__name

    @property
    def size(self) -> int:
        self.__check_state_is_exists()
        return self.__state['size']

    @property
    def uncommitted_size(self) -> int:
        self.__check_state_is_exists()
        return self.__state['uncommitted_size']

    @property
    def root_hash(self) -> str:
        self.__check_state_is_exists()
        return self.__state['root_hash']

    @property
    def uncommitted_root_hash(self) -> str:
        self.__check_state_is_exists()
        return self.__state['uncommitted_root_hash']

    @property
    def seq_no(self) -> int:
        self.__check_state_is_exists()
        return self.__state['seqNo']

    async def reload(self):
        state = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/state',
            params={
                'name': self.name
            }
        )
        self.__state = state

    async def rename(self, new_name: str):
        await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/rename',
            params={
                'name': self.name,
                'new_name': new_name
            }
        )
        self.__name = new_name

    async def init(self, genesis: List[Transaction]) -> List[Transaction]:
        self.__state, txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/initialize',
            params={
                'name': self.name,
                'genesis_txns': genesis
            }
        )
        txns = [Transaction.from_value(txn) for txn in txns]
        return txns

    async def append(
            self, transactions: Union[List[Transaction], List[dict]], txn_time: Union[str, int] = None
    ) -> (int, int, List[Transaction]):
        transactions_to_append = []
        for txn in transactions:
            if isinstance(txn, Transaction):
                transactions_to_append.append(txn)
            elif isinstance(txn, dict):
                transactions_to_append.append(Transaction.create(txn))
            else:
                raise RuntimeError('Unexpected transaction type')
        transactions_with_meta = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/append_txns_metadata',
            params={
                'name': self.name,
                'txns': transactions_to_append,
                'txn_time': txn_time
            }
        )
        self.__state, start, end, appended_txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/append_txns',
            params={
                'name': self.name,
                'txns': transactions_with_meta,
            }
        )
        return start, end, Transaction.from_value(appended_txns)

    async def commit(self, count: int) -> (int, int, List[Transaction]):
        self.__state, start, end, committed_txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/commit_txns',
            params={
                'name': self.name,
                'count': count,
            }
        )
        return start, end, Transaction.from_value(committed_txns)

    async def discard(self, count: int):
        self.__state = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/discard_txns',
            params={
                'name': self.name,
                'count': count,
            }
        )

    async def merkle_info(self, seq_no: int) -> MerkleInfo:
        merkle_info = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/merkle_info',
            params={
                'name': self.name,
                'seqNo': seq_no,
            }
        )
        return MerkleInfo(
            root_hash=merkle_info['rootHash'],
            audit_path=merkle_info['auditPath'],
        )

    async def audit_proof(self, seq_no: int) -> AuditProof:
        proof = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/audit_proof',
            params={
                'name': self.name,
                'seqNo': seq_no,
            }
        )
        return AuditProof(
            root_hash=proof['rootHash'],
            audit_path=proof['auditPath'],
            ledger_size=proof['ledgerSize']
        )

    async def reset_uncommitted(self):
        self.__state = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/reset_uncommitted',
            params={
                'name': self.name,
            }
        )

    async def get_transaction(self, seq_no: int) -> Transaction:
        txn = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_by_seq_no',
            params={
                'name': self.name,
                'seqNo': seq_no
            }
        )
        txn = Transaction.from_value(txn)
        assert isinstance(txn, Transaction)
        return txn

    async def get_uncommitted_transaction(self, seq_no: int) -> Transaction:
        txn = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_by_seq_no_uncommitted',
            params={
                'name': self.name,
                'seqNo': seq_no
            }
        )
        txn = Transaction.from_value(txn)
        assert isinstance(txn, Transaction)
        return txn

    async def get_last_transaction(self) -> Transaction:
        txn = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_last_txn',
            params={
                'name': self.name
            }
        )
        txn = Transaction.from_value(txn)
        assert isinstance(txn, Transaction)
        return txn

    async def get_last_committed_transaction(self) -> Transaction:
        txn = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_last_committed_txn',
            params={
                'name': self.name
            }
        )
        txn = Transaction.from_value(txn)
        assert isinstance(txn, Transaction)
        return txn

    async def get_all_transactions(self) -> List[Transaction]:
        txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_all_txns',
            params={
                'name': self.name
            }
        )
        txns = [t[1] for t in txns]
        txns = Transaction.from_value(txns)
        assert isinstance(txns, list)
        return txns

    async def get_uncommitted_transactions(self) -> List[Transaction]:
        txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_uncommitted_txns',
            params={
                'name': self.name
            }
        )
        txns = Transaction.from_value(txns)
        assert isinstance(txns, list)
        return txns

    def __check_state_is_exists(self):
        if self.__state is None:
            raise SiriusContextError('Load state of Microledger at First!')


class MicroledgerList:

    def __init__(self, api: AgentRPC):
        self.__api = api
        self.instances = {}

    async def create(self, name: str, genesis: Union[List[Transaction], List[dict]]) -> (Microledger, List[Transaction]):
        genesis_txns = []
        for txn in genesis:
            if isinstance(txn, Transaction):
                genesis_txns.append(txn)
            elif isinstance(txn, dict):
                genesis_txns.append(Transaction.create(txn))
            else:
                raise RuntimeError('Unexpected transaction type')
        instance = Microledger(name, self.__api)
        txns = await instance.init(genesis_txns)
        self.instances[name] = instance
        return instance, txns

    async def ledger(self, name: str) -> Microledger:
        if name not in self.instances:
            await self.__check_is_exists(name)
            instance = Microledger(name, self.__api)
            self.instances[name] = instance
        return self.instances[name]

    async def reset(self, name: str):
        await self.__check_is_exists(name)
        await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/reset',
            params={
                'name': name
            }
        )
        if name in self.instances.keys():
            del self.instances[name]

    async def is_exists(self, name: str):
        is_exists = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/is_exists',
            params={
                'name': name
            }
        )
        return is_exists

    async def leaf_hash(self, txn: Union[Transaction, bytes]) -> bytes:
        if isinstance(txn, Transaction):
            data = json.dumps(txn, sort_keys=True, ensure_ascii=False, separators=(',', ':')).encode()
        elif isinstance(txn, bytes):
            data = txn
        else:
            raise RuntimeError('Unexpected transaction type')
        leaf_hash = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/leaf_hash',
            params={
                'data': data
            }
        )
        return leaf_hash

    async def list(self) -> List[LedgerMeta]:
        collection = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/list',
            params={
                'name': '*'
            }
        )
        return [LedgerMeta(**item) for item in collection]

    async def __check_is_exists(self, name: str):
        if name not in self.instances.keys():
            is_exists = await self.is_exists(name)
            if not is_exists:
                raise SiriusContextError('MicroLedger with name "" does not exists' % name)



from abc import ABC, abstractmethod
from typing import Any, Optional


class AbstractKeyValueStorage(ABC):

    @abstractmethod
    async def select_db(self, db_name: str):
        raise NotImplemented

    @abstractmethod
    async def set(self, key: str, value: Any):
        raise NotImplemented

    @abstractmethod
    async def get(self, key: str) -> Optional[Any]:
        raise NotImplemented

    @abstractmethod
    async def delete(self, key: str):
        raise NotImplemented
